package com.sun.isve.actor

import scala.actors.Actor
import scala.actors.Actor._
import _root_.java.net.Socket
import _root_.java.net.ServerSocket
import _root_.java.io.OutputStreamWriter
import _root_.java.io.InputStreamReader
import _root_.java.io.LineNumberReader

case class Idle(worker: ServerWorker)
case class Connection(socket: Socket, id: Int)

class ServerWorker(val id: Int, val dispatcher: Dispatcher) extends Actor
{
    def act()
    {
        loop
        {
            react
            {
                case Connection(socket, id) =>
                    handleConnection(socket)
                    socket.close()
                    dispatcher ! Idle(this)
            }
        }
    }

    override def hashCode(): Int = id

    override def equals(other: Any): Boolean =
        other match
        {
            case that: ServerWorker => this.id == that.id
            case _                  => false
        }

    def handleConnection(socket: Socket) =
    {
        val os = socket.getOutputStream
        val writer = new OutputStreamWriter(os)

        val is = socket.getInputStream
        val reader = new LineNumberReader(new InputStreamReader(is))

        reader.readLine()
        writer.write(Utility.getPage("ScalaServer", 512))
        writer.flush()
    }
}

class Dispatcher() extends Actor
{
    import scala.collection.mutable.{Map, ListBuffer}
    import _root_.java.util.Random

    val idleWorkers = new ListBuffer[ServerWorker]
    val busyWorkers = Map[Int, ServerWorker]()
    val rng = new Random()

    for (i <- 1 to Runtime.getRuntime().availableProcessors() * 4 + 1)
    {
        val w = new ServerWorker(i, this)
        w.start()
        idleWorkers += w
    }

    def act()
    {
        loop
        {
            react
            {
                case Idle(worker) =>
                    busyWorkers -= worker.id
                    idleWorkers += worker

                case conn: Connection =>
                    val worker =
                        if (idleWorkers.length == 0)
                            busyWorkers.get(rng.nextInt(busyWorkers.size)).get
                        else
                        {
                            val w = idleWorkers.remove(0)
                            busyWorkers += w.id -> w
                            w
                        }

                    worker ! conn
            }
        }

    }
}

class Server()
{
    val name: String = "Main"

    def run() =
    {
        val socket = new ServerSocket(Utility.PORT)
        val dispatcher = new Dispatcher()
        var i = 0

        dispatcher.start()

        while (true)
        {
            val clientConn = socket.accept()
            i += 1
            dispatcher ! Connection(clientConn, i)
        }
    }
}

object ScalaWebserver extends Application
{
    println("This is ScalaWebserver running on port " + Utility.PORT);
    new Server().run()
}
