package com.sun.isve.nio

import scala.actors.Actor
import scala.actors.Actor._

import _root_.java.nio.channels.Selector
import _root_.java.nio.channels.spi.SelectorProvider
import _root_.java.nio.channels.SelectionKey
import _root_.java.nio.channels.Selector
import _root_.java.nio.channels.ServerSocketChannel
import _root_.java.nio.channels.SocketChannel
import _root_.java.net.InetSocketAddress;
import _root_.java.nio.ByteBuffer

case class Idle(worker: ServerWorker)
case class Read(channel:SocketChannel, id: Int)

class ServerWorker(val id: Int, val dispatcher: Dispatcher) extends Actor
{
    def act()
    {
        loop
        {
            react
            {
                case Read(channel, id) =>
                    channel.write(ByteBuffer.wrap((Utility.getPage("ScalaNioServer", 512).getBytes())))
                    channel.close()
                    
                    dispatcher ! Idle(this)
            }
        }
    }

    override def hashCode(): Int = id

    override def equals(other: Any): Boolean =
        other match
        {
            case that: ServerWorker => this.id == that.id
            case _                  => false
        }
}


class Dispatcher() extends Actor
{
    import scala.collection.mutable.{Map, ListBuffer}
    import _root_.java.util.Random

    val idleWorkers = new ListBuffer[ServerWorker]
    val busyWorkers = Map[Int, ServerWorker]()
    val rng = new Random()

    for (i <- 1 to Runtime.getRuntime().availableProcessors() * 4 + 1)
    {
        val w = new ServerWorker(i, this)
        w.start()
        idleWorkers += w
    }

    def act()
    {
        loop
        {
            react
            {
                case read: Read =>
                    getWorker() ! read

                case Idle(worker) =>
                    busyWorkers -= worker.id
                    idleWorkers += worker
            }
        }

    }

    def getWorker(): ServerWorker = {
        if (idleWorkers.length == 0)
            busyWorkers.get(rng.nextInt(busyWorkers.size)).get
        else
        {
            val w = idleWorkers.remove(0)
            busyWorkers += w.id -> w
            w
        }
    }

}

class SelectingRunnable()
{
    val selector = initSelector(Utility.PORT);

    def run() =
    {
        val dispatcher = new Dispatcher()
        var i = 0

        dispatcher.start()

        while (true)
        {
            selector.select()

            val selectedKeysItr = selector.selectedKeys().iterator()

            while (selectedKeysItr.hasNext()) {
                val key = selectedKeysItr.next().asInstanceOf[SelectionKey]
                selectedKeysItr.remove();

                if (!key.isValid()) {
                    continue;
                }

                if (key.isAcceptable()) {
                    accept(key)
                } else if (key.isReadable()) {
                    i += 1
                    dispatcher ! Read(key.channel().asInstanceOf[SocketChannel], i)
                    key.cancel
                }
            }
        }
    }

    def accept(key:SelectionKey) = {
        val serverSocketChannel = key.channel().asInstanceOf[ServerSocketChannel]
        val socketChannel = serverSocketChannel.accept()

        socketChannel.configureBlocking(false);
        socketChannel.register(selector, SelectionKey.OP_READ);
    }



    def initSelector(port:int): Selector = {
        val socketSelector = SelectorProvider.provider().openSelector()
        val serverChannel = ServerSocketChannel.open()
        val isa = new InetSocketAddress(port)

        serverChannel.configureBlocking(false)
        serverChannel.socket().bind(isa)
        serverChannel.register(socketSelector, SelectionKey.OP_ACCEPT)

        return socketSelector
    }
}

object ScalaNioServer extends Application
{
    println("This is ScalaNioServer running on port " + Utility.PORT);
    new SelectingRunnable().run()
}
